#include <string.h>
#include <spu_mfcio.h>
#include "common.h"

// Masses of all bodies
float mass[6][SPU_BODIES] CACHE_ALIGNED;
// Position vectors for all bodies
VEC3D pos[6][SPU_BODIES] CACHE_ALIGNED;

// Pointer into mass array to section of bodies this SPU is responsible for
float *own_mass;
// Pointer into pos array to section of bodies this SPU is responsible for
VEC3D *own_pos;
// Velocity vectors for bodies this SPU is responsible for
VEC3D own_vel[SPU_BODIES] CACHE_ALIGNED;
// Acceleration vectors for bodies this SPU is responsible for
VEC3D own_acc[SPU_BODIES];

CONTROL_BLOCK cb QWORD_ALIGNED;

// Which section of bodies this SPU is responsible for (0-5)
int id;

// Updates acceleration vectors taking into account only interactions between
// the bodies this SPU is responsible for
void
process_own()
{
  for (int i = 0; i < SPU_BODIES - 1; i++) {
    for (int j = i + 1; j < SPU_BODIES; j++) {
      VEC3D d;
      float t;

      // Calculate displacement from i to j
      d.x = own_pos[j].x - own_pos[i].x;
      d.y = own_pos[j].y - own_pos[i].y;
      d.z = own_pos[j].z - own_pos[i].z;

      // Calculate 1 / dist^2
      t = 1 / (fsqrf(d.x) + fsqrf(d.y) + fsqrf(d.z));

      // Calculate components of t along the displacement vector
      t *= (float)sqrt(t);
      d.x *= t;
      d.y *= t;
      d.z *= t;

      // Update acceleration of i 
      own_acc[i].x += d.x * own_mass[j];
      own_acc[i].y += d.y * own_mass[j];
      own_acc[i].z += d.z * own_mass[j];

      // Update acceleration of j
      own_acc[j].x += -d.x * own_mass[i];
      own_acc[j].y += -d.y * own_mass[i];
      own_acc[j].z += -d.z * own_mass[i];
    }
  }
}

// Updates acceleration vectors taking into account interactions with the given
// section of bodies
void
process_other(VEC3D *other_pos, float *other_mass)
{
  for (int i = 0; i < SPU_BODIES; i++) {
    for (int j = 0; j < SPU_BODIES; j++) {
      VEC3D d;
      float t;

      // Calculate displacement from i to j
      d.x = other_pos[j].x - own_pos[i].x;
      d.y = other_pos[j].y - own_pos[i].y;
      d.z = other_pos[j].z - own_pos[i].z;

      // Calculate 1 / dist^2
      t = 1 / (fsqrf(d.x) + fsqrf(d.y) + fsqrf(d.z));

      // Calculate components of t along the displacement vector
      t *= (float)sqrt(t);
      d.x *= t;
      d.y *= t;
      d.z *= t;

      // Update acceleration of i 
      own_acc[i].x += d.x * other_mass[j];
      own_acc[i].y += d.y * other_mass[j];
      own_acc[i].z += d.z * other_mass[j];
    }
  }
}

int
main(uint64_t speid, uint64_t argp, uint64_t envp)
{
  id = envp;
  own_mass = mass[id];
  own_pos = pos[id];

  // DMA in control block and wait for completion
  mfc_get(&cb,
          argp,
          sizeof(cb),
          0,
          0, 0);
  mfc_write_tag_mask(1 << 0);
  mfc_read_tag_status_all();

  // DMA in masses of all bodies and initial positions/velocities of bodies
  // this SPU is responsible for
  mfc_get(mass,
          cb.mass_addr,
          sizeof(mass),
          0,
          0, 0);
  mfc_get(own_pos,
          cb.pos_addr + id * sizeof(pos[0]),
          sizeof(pos[0]),
          0,
          0, 0);
  mfc_get(own_vel,
          cb.vel_addr + id * sizeof(own_vel),
          sizeof(own_vel),
          0,
          0, 0);
  mfc_read_tag_status_all();

  while (TRUE) {
    // Wait for message from PPU indicating SPU can compute next step
    // (0 indicates SPU should stop)
    if (spu_read_in_mbox() != 0) {
      break;
    }

    // DMA in new positions of bodies other SPUs are responsible for
    if (id != 0) {
      mfc_get(pos,
              cb.pos_addr,
              id * sizeof(pos[0]),
              0,
              0, 0);
    }

    if (id != 5) {
      mfc_get(pos[id + 1],
              cb.pos_addr + (id + 1) * sizeof(pos[0]),
              (5 - id) * sizeof(pos[0]),
              0,
              0, 0);
    }

    // Wait for all positions to be transferred
    mfc_read_tag_status_all();

    // Notify PPU that positions have been read
    spu_write_out_mbox(0);

    // Clear acceleration vectors
    memset(own_acc, 0, sizeof(own_acc));

    // Process interactions between the bodies this SPU is responsible for
    process_own();

    // Process interactions with other bodies
    for (int i = 0; i < 6; i++) {
      if (i != id) {
        process_other(pos[i], mass[i]);
      }
    }

    // Update positions and velocities
    for (int i = 0; i < SPU_BODIES; i++) {
      own_pos[i].x += own_vel[i].x * TIMESTEP;
      own_pos[i].y += own_vel[i].y * TIMESTEP;
      own_pos[i].z += own_vel[i].z * TIMESTEP;

      own_vel[i].x += own_acc[i].x * TIMESTEP;
      own_vel[i].y += own_acc[i].y * TIMESTEP;
      own_vel[i].z += own_acc[i].z * TIMESTEP;
    }

    // Wait for message from PPU indicating it is safe to write back new
    // positions
    spu_read_in_mbox();

    // Write back new positions for the bodies this SPU is responsible for and
    // wait for DMA completion
    mfc_put(own_pos,
            cb.pos_addr + id * sizeof(pos[0]),
            sizeof(pos[0]),
            0,
            0, 0);
    mfc_read_tag_status_all();

    // Notify PPU that new positions have been written
    spu_write_out_mbox(0);
  }

  return 0;
}
